/*
 * Copyright (c) 2017, Harald Kuhr
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * * Neither the name of the copyright holder nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package com.twelvemonkeys.imageio.plugins.webp;

/**
 * A special VP8* RIFF chunk, used as the first chunk of a WebP file.
 */
final class VP8xChunk extends RIFFChunk {
    int width;
    int height;

    boolean isLossless;

    boolean containsICCP;
    boolean containsALPH;
    boolean containsEXIF;
    boolean containsXMP_;
    boolean containsANIM;

    VP8xChunk(int fourCC, long length, long offset) {
        super(fourCC, length, offset);
    }

    @Override
    public String toString() {
        return super.toString() + "[" +
                "width=" + width +
                ", height=" + height +
                ", lossless=" + (isLossless ? "RGB" : "")
                + (containsALPH ? "A" : (isLossless ? "" : "false")) +
                ", flags=" +
                (containsICCP ? "I" : "") +
                (containsALPH ? "L" : "") +
                (containsEXIF ? "E" : "") +
                (containsXMP_ ? "X" : "") +
                (containsANIM ? "A" : "") +
                ']';
    }
}
